﻿Option Strict On
Option Infer On

Partial Class Cluster(Of TCell As Cell(Of TCell, TConnection), TConnection As Connection(Of TCell, TConnection))

    Protected Class NeighbourInfo

        Sub New(ByVal Connection As TConnection, ByVal OtherCell As TCell)
            Me.Connection = Connection
            Me.OtherCell = OtherCell
        End Sub

        Public ReadOnly Connection As TConnection

        Public ReadOnly OtherCell As TCell

    End Class

    Protected Class NeighbourInfoExt
        Inherits NeighbourInfo

        Sub New(ByVal NeighbourInfo As NeighbourInfo, ByVal Delta As Double)
            Call MyBase.New(Connection:=NeighbourInfo.Connection, OtherCell:=NeighbourInfo.OtherCell)
            Me.Delta = Delta
        End Sub

        Public ReadOnly Delta As Double

    End Class

    Protected Class NeighbourhoodInfo

        Public Sub New(ByVal ActiveNeighbours() As NeighbourInfoExt, ByVal ActiveValue As Double, ByVal Average As Double)
            Me.ActiveNeighbours = ActiveNeighbours
            Me.ActiveValue = ActiveValue
            Me.Average = Average
        End Sub

        Public ReadOnly ActiveNeighbours() As NeighbourInfoExt

        Public ReadOnly ActiveValue As Double

        Public ReadOnly Average As Double

    End Class

    ''' <summary>
    ''' Returns all neighbours of the given cell based on its connections
    ''' </summary>
    Protected Shared Function FindNeighbours(ByVal cell As TCell) As NeighbourInfo()
        Dim Connections = cell.Parent.FindConnections(cell)
        Dim Neighbours(0 To Connections.Length - 1) As NeighbourInfo
        For i As Integer = 0 To Connections.Length - 1
            Dim Connection As TConnection = Connections(i)
            Dim OtherCell As TCell = Connection.OtherCell(cell)
            Neighbours(i) = New NeighbourInfo(Connection:=Connection, OtherCell:=OtherCell)
        Next i

        Return Neighbours
    End Function

    ''' <summary>
    ''' Returns all neighbours of the given cell based on its connections if predicate returns true for a given neighbour
    ''' </summary>
    Protected Shared Function FindNeighbours(ByVal cell As TCell, ByVal predicate As Func(Of TConnection, Boolean)) As NeighbourInfo()
        Dim Connections() As TConnection = cell.Parent.FindConnections(cell)
        Dim Neighbours As New List(Of NeighbourInfo)(capacity:=Connections.Length)
        For i As Integer = 0 To Connections.Length - 1
            Dim Connection As TConnection = Connections(i)
            If Not predicate(Connection) Then Continue For
            Dim OtherCell As TCell = Connection.OtherCell(cell)
            Neighbours.Add(New NeighbourInfo(Connection:=Connection, OtherCell:=OtherCell))
        Next i

        Return Neighbours.ToArray
    End Function

    ''' <summary>
    ''' Returns all neighbours of the given cell based on its connections if predicate returns true for a given neighbour
    ''' </summary>
    Protected Shared Function FindNeighbours(ByVal cell As TCell, ByVal predicate As Func(Of TCell, Boolean)) As NeighbourInfo()
        Dim Connections() As TConnection = cell.Parent.FindConnections(cell)
        Dim Neighbours As New List(Of NeighbourInfo)(capacity:=Connections.Length)
        For i As Integer = 0 To Connections.Length - 1
            Dim Connection As TConnection = Connections(i)
            Dim OtherCell As TCell = Connection.OtherCell(cell)
            If Not predicate(OtherCell) Then Continue For
            Neighbours.Add(New NeighbourInfo(Connection:=Connection, OtherCell:=OtherCell))
        Next i

        Return Neighbours.ToArray
    End Function




End Class
