'Cellular Automata Class Library
'Copyright (C) 2010, Jrg Meier
'see folder [license] for details.

Option Strict On

<System.CLSCompliant(True)> _
Public MustInherit Class Cluster(Of TCell As Cell(Of TCell, TConnection), TConnection As Connection(Of TCell, TConnection))
    Implements ICluster

    Sub New()
        Me._Cells = New ObservableCollection(Of TCell)
        Me._Connections = New ObservableCollection(Of TConnection)
    End Sub

    Private WithEvents _Cells As ObservableCollection(Of TCell)
    Public ReadOnly Property Cells() As IList(Of TCell)
        Get
            Return _Cells
        End Get
    End Property
    Private Sub _Cells_OnBeforeClearItems() Handles _Cells.OnBeforeClearItems
        Call _Connections.Clear()
        Call _ConnectionCache.Clear()
    End Sub
    Private Sub _Cells_OnBeforeInsertItem(ByVal index As Integer, ByVal item As TCell) Handles _Cells.OnBeforeInsertItem
        'die Indices setzen!
        For i As Integer = index To _Cells.Count - 1
            Call _Cells.Item(i).SetIndex(i + 1)
        Next
        Call item.SetIndex(index)
    End Sub
    Private Sub _Cells_OnBeforeRemoveItem(ByVal index As Integer) Handles _Cells.OnBeforeRemoveItem
        'die Connections Updaten!
        Dim CellConnections As TConnection() = FindConnections(_Cells(index))
        For Each Connection As TConnection In Connections
            Call _Connections.Remove(Connection)
        Next
        Call _ConnectionCache.Remove(_Cells(index))

        'die Indices setzen!
        For i As Integer = index To _Cells.Count - 1
            Call _Cells.Item(i).SetIndex(i - 1)
        Next
    End Sub
    Private Sub _Cells_OnBeforeSetItem(ByVal index As Integer, ByVal item As TCell) Handles _Cells.OnBeforeSetItem
        Call _Cells_OnBeforeRemoveItem(index)

        'Index setzen
        Call item.SetIndex(index)
    End Sub

    Private ReadOnly Property CellCount() As Integer Implements ICluster.CellCount
        Get
            Return _Cells.Count
        End Get
    End Property

    Private Property Cell(ByVal index As Integer) As ICell Implements ICluster.Cell
        Get
            Return _Cells.Item(index:=index)
        End Get
        Set(ByVal value As ICell)
            _Cells.Item(index:=index) = CType(value, TCell)
        End Set
    End Property


    Private WithEvents _Connections As ObservableCollection(Of TConnection)
    Public ReadOnly Property Connections() As IList(Of TConnection)
        Get
            Return _Connections
        End Get
    End Property
    Private Sub _Connections_OnBeforeClearItems() Handles _Connections.OnBeforeClearItems
        _ConnectionCache.Clear()
    End Sub
    Private Sub _Connections_OnBeforeInsertItem(ByVal index As Integer, ByVal item As TConnection) Handles _Connections.OnBeforeInsertItem
        _ConnectionCache.Remove(item.Cell1)
        _ConnectionCache.Remove(item.Cell2)

        'die Indices setzen!
        For i As Integer = index To _Connections.Count - 1
            Call _Connections.Item(i).SetIndex(i + 1)
        Next
        Call item.SetIndex(index)
    End Sub
    Private Sub _Connections_OnBeforeRemoveItem(ByVal index As Integer) Handles _Connections.OnBeforeRemoveItem
        Dim item As TConnection = _Connections(index)
        _ConnectionCache.Remove(item.Cell1)
        _ConnectionCache.Remove(item.Cell2)

        'die Indices setzen!
        For i As Integer = index To _Connections.Count - 1
            Call _Connections.Item(i).SetIndex(i - 1)
        Next
    End Sub
    Private Sub _Connections_OnBeforeSetItem(ByVal index As Integer, ByVal item As TConnection) Handles _Connections.OnBeforeSetItem
        _ConnectionCache.Remove(item.Cell1)
        _ConnectionCache.Remove(item.Cell2)

        'Index setzen
        Call item.SetIndex(index)
    End Sub

    Private Function ICluster_Connections() As IEnumerable(Of IConnection) Implements ICluster.FindConnections
        Return _Connections.Cast(Of IConnection)()
    End Function


    Private _ConnectionCache As New Dictionary(Of Cell(Of TCell, TConnection), TConnection())

    Public Function FindConnections(ByVal Cell As Cell(Of TCell, TConnection)) As TConnection()
        Dim Result() As TConnection = Nothing
        Dim Found As Boolean = _ConnectionCache.TryGetValue(Cell, Result)
        If Found Then
            Return Result
        Else
            Dim List As New List(Of TConnection)(capacity:=_Connections.Count)
            For Each Connection As TConnection In _Connections
                If Connection.Cell1 Is Cell OrElse Connection.Cell2 Is Cell Then
                    Call List.Add(Connection)
                End If
            Next
            Result = List.ToArray
            Call _ConnectionCache.Add(Cell, Result)
            Return Result
        End If
    End Function
    Private Function ICluster_FindConnections(ByVal cell As ICell) As IEnumerable(Of IConnection) Implements ICluster.FindConnections
        Dim result As TConnection() = FindConnections(CType(cell, Global.CellularAutomata.Cell(Of TCell, TConnection)))
        Return result.Cast(Of IConnection)()
    End Function

    Protected Sub RegisterCellConnections(ByVal Cell As Cell(Of TCell, TConnection), ByVal Connections() As TConnection)
        If _ConnectionCache.ContainsKey(Cell) Then
            _ConnectionCache.Item(Cell) = Connections
        Else
            _ConnectionCache.Add(Cell, Connections)
        End If
    End Sub


    Private _StepIndex As Integer
    Public ReadOnly Property StepIndex() As Integer Implements ICluster.StepIndex
        Get
            Return _StepIndex
        End Get
    End Property


    Private _SimulationTime As Double = 0

    ''' <summary>
    ''' Gives the total Timespan elapsed for the Cluster
    ''' </summary>
    Public ReadOnly Property SimulationTime() As Double Implements ICluster.SimulationTime
        Get
            Return _SimulationTime
        End Get
    End Property



    Public Function DoStep(ByVal e As DoStepArgs) As StepResult Implements ICluster.DoStep
        Dim result As StepResult = OnDoStep(e:=e)

        If result.StepSuccessful Then
            _StepIndex += 1
            _SimulationTime += e.TimeStepSize
        End If

        Return result
    End Function

    Protected MustOverride Function OnDoStep(ByVal e As DoStepArgs) As StepResult


End Class
