'Cellular Automata Class Library
'Copyright (C) 2010, Jrg Meier
'see folder [license] for details.

Option Strict On
Option Infer On

Imports CellularAutomata

Namespace Specialized

    Partial Class SpecializedCluster(Of TCell As Cell(Of TCell, TConnection), TConnection As Connection(Of TCell, TConnection))

        <System.ComponentModel.EditorBrowsable(ComponentModel.EditorBrowsableState.Advanced)>
        Public Delegate Function Rectangular1DCellFactory(ByVal cluster As Cluster(Of TCell, TConnection), ByVal i As Integer) As TCell

        <System.ComponentModel.EditorBrowsable(ComponentModel.EditorBrowsableState.Advanced)>
        Public Delegate Function Rectangular2DCellFactory(ByVal cluster As Cluster(Of TCell, TConnection), ByVal i As Integer, j As Integer) As TCell

        <System.ComponentModel.EditorBrowsable(ComponentModel.EditorBrowsableState.Advanced)>
        Public Delegate Function Rectangular3DCellFactory(ByVal cluster As Cluster(Of TCell, TConnection), ByVal i As Integer, j As Integer, k As Integer) As TCell


        Friend MustInherit Class RectangularClusterFactory
            Private Sub New()
            End Sub

            ''' <summary>
            ''' Creates a 1D rectangular grid of cells and adds it to the given cluster
            ''' </summary>
            ''' <param name="cluster"></param>
            ''' <param name="CellCount"></param>
            ''' <param name="CellFactory"></param>
            ''' <param name="ConnectionFactory"></param>
            ''' <returns></returns>
            ''' <remarks></remarks>
            Friend Shared Function AddRectangularGridCluster( _
                ByVal cluster As SpecializedCluster(Of TCell, TConnection), _
                ByVal CellCount As Integer, _
                ByVal CellFactory As Rectangular1DCellFactory, _
                ByVal ConnectionFactory As ConnectionFactory) As TCell()

                Dim CellFactory2D = Function(c As Cluster(Of TCell, TConnection), i As Integer, j As Integer) As TCell
                                        Return CellFactory.Invoke(c, i)
                                    End Function

                Dim arr As TCell(,) = AddRectangularGridCluster(cluster:=cluster,
                                                                CellCountX:=CellCount, CellCountY:=1,
                                                                CellFactory:=CellFactory2D,
                                                                ConnectionFactory:=ConnectionFactory,
                                                                AddSecondLevelConnections:=False)

                Dim result(0 To CellCount - 1) As TCell
                For i As Integer = 0 To CellCount - 1
                    result(i) = arr(i, 0)
                Next
                Return result
            End Function

            ''' <summary>
            ''' Creates a 2D rectangular grid of cells and adds it to the given cluster
            ''' </summary>
            ''' <param name="cluster"></param>
            ''' <param name="CellCountX"></param>
            ''' <param name="CellCountY"></param>
            ''' <param name="CellFactory"></param>
            ''' <param name="ConnectionFactory"></param>
            ''' <param name="AddSecondLevelConnections"></param>
            ''' <returns></returns>
            ''' <remarks></remarks>
            Friend Shared Function AddRectangularGridCluster( _
                ByVal cluster As SpecializedCluster(Of TCell, TConnection), _
                ByVal CellCountX As Integer, _
                ByVal CellCountY As Integer, _
                ByVal CellFactory As Rectangular2DCellFactory, _
                ByVal ConnectionFactory As ConnectionFactory, _
                ByVal AddSecondLevelConnections As Boolean) As TCell(,)


                Dim CellFactory3D = Function(c As Cluster(Of TCell, TConnection), i As Integer, j As Integer, k As Integer) As TCell
                                        Return CellFactory.Invoke(c, i, j)
                                    End Function

                Dim arr As TCell(,,) = AddRectangularGridCluster(cluster:=cluster,
                                                                 CellCountX:=CellCountX, CellCountY:=CellCountY, CellCountZ:=1,
                                                                 CellFactory:=CellFactory3D,
                                                                 ConnectionFactory:=ConnectionFactory,
                                                                 AddSecondLevelConnections:=AddSecondLevelConnections,
                                                                 AddTertiaryLevelConnections:=False)

                Dim result(0 To CellCountX - 1, 0 To CellCountY - 1) As TCell
                For i As Integer = 0 To CellCountX - 1
                    For j As Integer = 0 To CellCountX - 1
                        result(i, j) = arr(i, j, 0)
                    Next
                Next
                Return result
            End Function


            ''' <summary>
            ''' Creates a 2D rectangular grid of cells and adds it to the given cluster
            ''' </summary>
            ''' <param name="cluster"></param>
            ''' <param name="CellCountX"></param>
            ''' <param name="CellCountY"></param>
            ''' <param name="CellCountZ"></param>
            ''' <param name="CellFactory"></param>
            ''' <param name="ConnectionFactory"></param>
            ''' <param name="AddSecondLevelConnections"></param>
            ''' <param name="AddTertiaryLevelConnections"></param>
            ''' <returns></returns>
            ''' <remarks></remarks>
            Friend Shared Function AddRectangularGridCluster( _
                ByVal cluster As SpecializedCluster(Of TCell, TConnection), _
                ByVal CellCountX As Integer, _
                ByVal CellCountY As Integer, _
                ByVal CellCountZ As Integer, _
                ByVal CellFactory As Rectangular3DCellFactory, _
                ByVal ConnectionFactory As ConnectionFactory, _
                ByVal AddSecondLevelConnections As Boolean, ByVal AddTertiaryLevelConnections As Boolean) As TCell(,,)

                Const AddPrimaryLevelConnections As Boolean = True

                'create all cells
                Dim Cells(0 To CellCountX - 1, 0 To CellCountY - 1, 0 To CellCountZ - 1) As TCell
                For i As Integer = Cells.GetLowerBound(0) To Cells.GetUpperBound(0)
                    For j As Integer = Cells.GetLowerBound(1) To Cells.GetUpperBound(1)
                        For k As Integer = Cells.GetLowerBound(2) To Cells.GetUpperBound(2)
                            Dim cell As TCell = CellFactory.Invoke(cluster:=cluster, i:=i, j:=j, k:=k)

                            Cells(i, j, k) = cell
                            If cell IsNot Nothing Then
                                Call cluster.Cells.Add(cell)
                            End If
                        Next
                    Next
                Next

                'create a array of connection-lists
                Dim CellConnections(0 To Cells.Length - 1) As List(Of TConnection)
                For i As Integer = 0 To CellConnections.Length - 1
                    CellConnections(i) = New List(Of TConnection)
                Next

                'create all needed connections
                For i As Integer = Cells.GetLowerBound(0) To Cells.GetUpperBound(0)
                    For j As Integer = Cells.GetLowerBound(1) To Cells.GetUpperBound(1)
                        For k As Integer = Cells.GetLowerBound(2) To Cells.GetUpperBound(2)
                            Dim Connections As New List(Of TConnection)

                            If AddPrimaryLevelConnections Then
                                AddConnection(Connections, Cells, i, j, k, +1, 0, 0, ConnectionTypes.Primary, ConnectionFactory)
                                AddConnection(Connections, Cells, i, j, k, 0, +1, 0, ConnectionTypes.Primary, ConnectionFactory)
                                AddConnection(Connections, Cells, i, j, k, 0, 0, +1, ConnectionTypes.Primary, ConnectionFactory)
                            End If

                            If AddSecondLevelConnections Then
                                AddConnection(Connections, Cells, i, j, k, +1, +1, 0, ConnectionTypes.Secondary, ConnectionFactory)
                                AddConnection(Connections, Cells, i, j, k, -1, +1, 0, ConnectionTypes.Secondary, ConnectionFactory)
                                AddConnection(Connections, Cells, i, j, k, +1, 0, +1, ConnectionTypes.Secondary, ConnectionFactory)
                                AddConnection(Connections, Cells, i, j, k, -1, 0, +1, ConnectionTypes.Secondary, ConnectionFactory)
                                AddConnection(Connections, Cells, i, j, k, 0, +1, +1, ConnectionTypes.Secondary, ConnectionFactory)
                                AddConnection(Connections, Cells, i, j, k, 0, -1, +1, ConnectionTypes.Secondary, ConnectionFactory)
                            End If

                            If AddTertiaryLevelConnections Then
                                AddConnection(Connections, Cells, i, j, k, +1, +1, +1, ConnectionTypes.Tertiary, ConnectionFactory)
                                AddConnection(Connections, Cells, i, j, k, -1, +1, +1, ConnectionTypes.Tertiary, ConnectionFactory)
                                AddConnection(Connections, Cells, i, j, k, +1, -1, +1, ConnectionTypes.Tertiary, ConnectionFactory)
                                AddConnection(Connections, Cells, i, j, k, -1, -1, +1, ConnectionTypes.Tertiary, ConnectionFactory)
                            End If

                            For Each Connection As TConnection In Connections
                                If Connection Is Nothing Then Continue For
                                If Connection.Cell1 Is Nothing OrElse Connection.Cell2 Is Nothing Then Continue For
                                Call cluster.Connections.Add(Connection)
                                Call CellConnections(Connection.Cell1.Index).Add(Connection)
                                Call CellConnections(Connection.Cell2.Index).Add(Connection)
                            Next Connection
                        Next
                    Next
                Next

                For i As Integer = 0 To cluster.Cells.Count - 1
                    Call cluster.RegisterCellConnections(cluster.Cells(i), CellConnections(i).ToArray)
                Next

                Return Cells
            End Function

            Private Shared Sub AddConnection(connections As ICollection(Of TConnection),
                                             cells(,,) As TCell,
                                             i As Integer, j As Integer, k As Integer,
                                             di As Integer, dj As Integer, dk As Integer,
                                             type As ConnectionTypes,
                                             ConnectionFactory As ConnectionFactory)

                If i < cells.GetLowerBound(0) OrElse i > cells.GetUpperBound(0) Then Return
                If j < cells.GetLowerBound(1) OrElse j > cells.GetUpperBound(1) Then Return
                If k < cells.GetLowerBound(2) OrElse k > cells.GetUpperBound(2) Then Return
                If (i + di) < cells.GetLowerBound(0) OrElse (i + di) > cells.GetUpperBound(0) Then Return
                If (j + dj) < cells.GetLowerBound(1) OrElse (j + dj) > cells.GetUpperBound(1) Then Return
                If (k + dk) < cells.GetLowerBound(2) OrElse (k + dk) > cells.GetUpperBound(2) Then Return

                Dim c1 = cells(i, j, k)
                Dim c2 = cells(i + di, j + dj, k + dk)
                If c1 Is Nothing Or c2 Is Nothing Then Return

                Dim connection As TConnection = ConnectionFactory(c1, c2, type)
                If connection IsNot Nothing Then
                    Call connections.Add(connection)
                End If
            End Sub

        End Class

    End Class

End Namespace
